#!/bin/sh
# Copyright 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This script builds the current version of the H3 documentation website
# and pushes it to production. This assumes that you have set the remote
# for gh-pages to the upstream (i.e. uber/h3) remote.

git checkout gh-pages
git pull
git checkout master
pushd website
./scripts/build-to-gh-pages.sh
git push
git checkout master
popd


