.playgroundContainer {
  margin-bottom: var(--ifm-leading);
  border-radius: var(--ifm-global-radius);
  box-shadow: var(--ifm-global-shadow-lw);
  overflow: hidden;
}

.playgroundHeader {
  letter-spacing: 0.08rem;
  padding: 0.75rem;
  text-transform: uppercase;
  font-weight: bold;
  background: var(--ifm-color-emphasis-200);
  color: var(--ifm-color-content);
  font-size: var(--ifm-code-font-size);
}

.playgroundHeader:first-of-type {
  background: var(--ifm-color-emphasis-600);
  color: var(--ifm-color-content-inverse);
}

.playgroundEditor {
  font: var(--ifm-code-font-size) / var(--ifm-pre-line-height)
    var(--ifm-font-family-monospace) !important;
  /* rtl:ignore */
  direction: ltr;
}

.playgroundPreview {
  padding: 1rem;
  background-color: var(--ifm-pre-background);

  /* H3 specific change for styling results*/
  max-height: 12em;
  overflow-x: hidden;
  overflow-y: auto;
}

/* H3 specific change for styling results*/
.playgroundPreview code {
  white-space: pre;
  border: 0;
  background-color: transparent;
}